(* This module is part of the bachelor thesis "Lambda-Interpreter", *)
(* written by Michael Brunner and supervised by Christian Sternagel, *)
(* University of Innsbruck, Institute of Computer Science *)
(* Date: 01.02.2009 *)
(* Copyright 2009 Michael Brunner, you may use this code or parts *)
(* of it as long as you keep this copyright notice in tact *)

(** minimalistic monad-type, primary used for quick definitions *)
module type MONAD_MINIMAL = sig
	type 'a t
	val bind: 'a t -> ('a -> 'b t) -> 'b t
	val unit: 'a -> 'a t
end;;

(** useful monad-type *)
module type MONAD = sig 
	type 'a t
	val (>>=): 'a t -> ('a -> 'b t) -> 'b t
	val (>>): 'a t -> 'b t -> 'b t
	val return: 'a -> 'a t
	val run: 'a t -> 'a
end;;

(** module-type to represent generic states used by state-monad *)
module type TYPEWRAPPER = sig 
	type t;; 
	val init: t;;
end;;


(** transform minimalistic monad into useful one *)
module Make (M: MONAD_MINIMAL) = struct 
	type 'a t = 'a M.t
	let (>>=) = M.bind
	let (>>) m n = m >>= fun _ -> n 
	let return = M.unit
end;;


(** identity-monad *)
module Identity = struct
	include Make (
		struct
			type 'a t = 'a;;
			let bind m f = f m;;
			let unit a = a;;
		end
	);;
  let run m = m;;
end;;

(** option monad *)
module Option = struct 
	module OptionM = struct
		type 'a t = 'a option;;
		let unit x = Some x;;
		let bind m f = match m with 
			| Some a -> f a
			| None -> None;;
	end
	include Make (OptionM);;
  let run = function 
		| Some a -> a 
		| None -> failwith "no value";;
  let fail = None;;
	let on_fail m h = match m with 
		| Some a -> Some a
		| None -> h;;
end;;

(** exception monad *)
module Exception = struct 
 	type 'a exc = Exception of string | Result of 'a;;
	module ExceptionM = struct
	  	type 'a t = 'a exc;;
			let unit x = Result x;;
			let bind m f = match m with 
				| Result a -> f a
				| Exception e -> Exception e;;
	end
	include Make (ExceptionM);;
	let raise e = Exception e;;
  let failed = function Exception (_) -> true | Result (_) -> false;;
  let trywith m h = match m with Exception e -> h e | Result r -> return r;;
  let run = function Exception (e) -> failwith e | Result (r) -> r;;
end;;

(** state monad *)
module State (S: TYPEWRAPPER) = struct 
	type state = S.t;;
	module StateM = struct
	  	type 'a t = state -> ('a * state);;
			let unit x = fun s -> (x, s);;
			let bind m f = fun s -> let (ri, si) = m s in 
				let (rf, sf) = (f ri) si in (rf, sf);;
	end
	include Make (StateM);;
  let read_state = fun s -> (s, s);;
  let write_state x = fun s -> ((), x);;
  let run m = fst (m S.init);;
  let run_with m s = fst (m s);;
end;;

(** IO monad *) 
module IO = struct 
	module IOM = struct 
		type 'a t = unit -> 'a;;
		let unit x = (fun () -> x);;
		let bind m f = (fun () -> f (m ()) ());;
	end
	include Make (IOM);;
  let write s = (fun () -> (print_string s));;
  let write_line s = (fun () -> (print_string (s ^ "\n")));;
  let read_line = (fun () -> try read_line () with End_of_file -> "\004");;
  let file_exists f = try let _ = open_in f in return true with _ -> return false;;
  let read_file f = 
		try 
		  let file = open_in f in
		  let rec read_line r= 
		    try 
				  let line = input_line file in 
			  	read_line (line::r)
		  	with 
	  			| _ -> r 
  		in return (List.rev (read_line []))
		with 
			| _ -> return [];;
  let write_file f contents = 
		try
			let file = open_out f in 
			let rec write_char = function 
				| [] -> (close_out file; return true)
				| c::cs -> (output_char file c; write_char cs) in
			write_char contents
		with 
			| _ -> return false;;
  let run m = m ();;
end;;

